implementation module ligretto

from   StdMisc import abort
import iTasks
import Math.Random		// for generating random numbers

get_random_int :: Task Int
get_random_int = get randomInt >>= \nr -> viewInformation "Random Int" [] nr

:: NrOfPlayers :== Int                            // 2 upto 4 players
:: Card          = { back     :: Color            // the backside color (belonging to player)
                   , front    :: Color            // the frontside color (for playing)
                   , nr       :: Int              // 1 upto 10
                   }
:: SideUp        = Front | Back                   // the side of the card that faces upwards
:: Color         = Red | Green | Blue | Yellow    // the four player / card colors
:: Player        = { color     :: Color           // the backside color of this player's cards
                   , row       :: Row             // the row of the player (nr_of_cards_in_row nr_of_players)
                   , ligretto  :: Pile            // the ligretto pile: 10 upto 0 (win) cards
                   , hand      :: Hand            // the hand cards
                   }
:: Pile        :== [Card]                         // a pile of cards
:: Row         :== [Card]                         // a row of (nr_of_cards_in_row nr_of_players) cards
:: Hand          = { conceal   :: Pile            // the concealed pile
                   , discard   :: Pile            // the discarded pile
                   }

nr_of_cards_in_row :: !NrOfPlayers -> Int
nr_of_cards_in_row 2 = 5
nr_of_cards_in_row 3 = 4
nr_of_cards_in_row 4 = 3
nr_of_cards_in_row n = abort ("ligretto.nr_of_cards_in_row: illegal integer argument (" +++ toString n +++ ").\n")

colors :: !NrOfPlayers -> [Color]
colors nr_of_players = take nr_of_players [Red,Green,Blue,Yellow]

initial_player_cards :: !NrOfPlayers !Color -> Pile
initial_player_cards nr_of_players back
	= [{back=back,front=color,nr=nr} \\ color <- colors nr_of_players, nr <- [1..10]]

shuffle :: ![a] !Int -> [a]
shuffle xs seed
	= fst (unzip (sortBy (\(_,r1) (_,r2) -> (r1 < r2)) (zip2 xs (genRandInt seed))))

initial_player :: !NrOfPlayers !Color !Int -> Player
initial_player nr_of_players back seed
	= { color = back, row = row, ligretto = ligretto, hand = { conceal = hand, discard = [] } }
where
	cards           = shuffle (initial_player_cards nr_of_players back) seed
	(row,rest)      = splitAt (nr_of_cards_in_row nr_of_players) cards
	(ligretto,hand) = splitAt 10 rest

pick_row_card :: !Int !Player -> (!Card,!Player)
pick_row_card row_nr player=:{row,ligretto}
| row_nr <= 0 || row_nr > length row
	= abort ("ligretto.pick_row_card: illegal integer argument (" <+++ row_nr <+++ ").\n")
| isEmpty ligretto
	= abort "ligretto.pick_row_card: trying to take card from empty ligretto.\n"
| otherwise
	= (card, {player & row = updateAt (row_nr-1) (hd ligretto) row, ligretto = tl ligretto})
where
	card = row !! (row_nr-1)

top_discard :: !Player -> Maybe Card
top_discard {hand={discard}}
| isEmpty discard   = Nothing
| otherwise         = Just (hd discard)

shuffle_hand :: !Int !Player -> Player
shuffle_hand seed player=:{hand=hand=:{conceal,discard}}
| isEmpty conceal   = {player & hand = { hand & conceal = shuffle discard seed
                                              , discard = []
                      }                }
| otherwise         = abort ("ligretto.shuffle_hand: not allowed to shuffle non-empty concealed pile.\n")

pick_discard :: !Player -> (!Card,!Player)
pick_discard player=:{hand=hand=:{conceal,discard}}
| isEmpty discard   = abort ("ligretto.pick_discard: no discarded card to pick.\n")
| otherwise         = (hd discard, {player & hand = { hand & discard = tl discard }})

swap_discards :: !Player -> Player
swap_discards player=:{hand=hand=:{conceal,discard}}
| isEmpty conceal   = abort ("ligretto:swap_discards: not allowed to take cards from an empty conceal pile.\n")
| otherwise         = { player & hand = { hand & conceal = rest
                                               , discard = reverse top3 ++ discard
                      }                 }
where
	(top3,rest)     = splitAt 3 conceal
